;; Generated package description from buffer-env.el  -*- no-byte-compile: t -*-
(define-package "buffer-env" "0.2" "Buffer-local process environments" '((emacs "27.1")) :commit "dc7342277f5094058f88f1a9a76cc8879f95d05a" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("processes" "tools") :url "https://github.com/astoff/buffer-env")
